#ifndef ITERATORS_PAIRIT_HH
#define ITERATORS_PAIRIT_HH

#include <iterator>
#include <cstddef>
#include <algorithm>
#include <stdexcept>
#include <new>
#include "rawmem.hh"

#ifdef _MSC_VER
#define NO_PARTIAL_SPECIALIZATION
#endif

namespace iterators
{
    
    namespace helper
    {
	typedef char Tiny;
	struct Small
	{
	    Tiny dummy[2];
	};
	struct Medium
	{
	    Small dummy[2];
	};
	struct Large
	{
	    Medium dummy[2];
	};
	struct Huge
	{
	    Large dummy[2];
	};

	template<unsigned>
	struct CategoryMap
	{
	    typedef void Type;
	};
	
	

//     Tiny categoryCheck(std::output_iterator_tag*);
	Small categoryCheck(std::input_iterator_tag*);
	Medium categoryCheck(std::forward_iterator_tag*);
	Large categoryCheck(std::bidirectional_iterator_tag*);
	Huge categoryCheck(std::random_access_iterator_tag*);


//     template<>
//     struct CategoryMap<sizeof(Tiny)>
//     {
// 	typedef std::output_iterator_tag Type;
//     };

	template<>
	struct CategoryMap<sizeof(Small)>
	{
	    typedef std::input_iterator_tag Type;
	};
   
	template<>
	struct CategoryMap<sizeof(Medium)>
	{
	    typedef std::forward_iterator_tag Type;
	};

	template<>
	struct CategoryMap<sizeof(Large)>
	{
	    typedef std::bidirectional_iterator_tag Type;
	};
	template<>
	struct CategoryMap<sizeof(Huge)>
	{
	    typedef std::random_access_iterator_tag Type;
	};

	template<typename Cat1,typename Cat2>
	struct CommonCategory
	{
	private:
	    enum
	    {categorySize=sizeof(helper::categoryCheck(false?(Cat1*)0:(Cat2*)0))
	    };
	public:
	    typedef typename CategoryMap<categorySize>::Type Type;
	};
	
	// specializations
#ifndef NO_PARTIAL_SPECIALIZATION
	template<typename Cat>
	struct CommonCategory<std::output_iterator_tag,Cat>
	{
	    typedef std::output_iterator_tag Type;
	};
	template<typename Cat>
	struct CommonCategory<Cat,std::output_iterator_tag>
	{
	    typedef std::output_iterator_tag Type;
	};
#else
	template<>
	struct CommonCategory<std::output_iterator_tag,std::forward_iterator_tag>
	{
	    typedef std::output_iterator_tag Type;
	};
	template<>
	struct CommonCategory<std::forward_iterator_tag,std::output_iterator_tag>
	{
	    typedef std::output_iterator_tag Type;
	};
	template<>
	struct CommonCategory<std::output_iterator_tag,std::bidirectional_iterator_tag>
	{
	    typedef std::output_iterator_tag Type;
	};
	template<>
	struct CommonCategory<std::bidirectional_iterator_tag,std::output_iterator_tag>
	{
	    typedef std::output_iterator_tag Type;
	};
	template<>
	struct CommonCategory<std::output_iterator_tag,std::random_access_iterator_tag>
	{
	    typedef std::output_iterator_tag Type;
	};
	template<>
	struct CommonCategory<std::random_access_iterator_tag,std::output_iterator_tag>
	{
	    typedef std::output_iterator_tag Type;
	};
#endif
	template<>
	struct CommonCategory<std::output_iterator_tag,std::output_iterator_tag>
	{
	    typedef std::output_iterator_tag Type;
	};
	template<>
	struct CommonCategory<std::input_iterator_tag,std::output_iterator_tag>
	{
	    // no Type, because error
	};
	template<>
	struct CommonCategory<std::output_iterator_tag,std::input_iterator_tag>
	{
	    // no Type, because error
	};
    }

    template<typename T,typename U>
    struct OwningRefPair;

    // a PlainPair holds a pair of values
    template<typename T,typename U>
    struct PlainPair
    {
	T first;
	U second;

	PlainPair(const T& t_,const U& u_):
	    first(t_),second(u_)
	{}

	template<typename X,typename Y>
	PlainPair& operator=(const OwningRefPair<X,Y>& other)
	{
	    first=other.first;
	    second=other.second;
	}
	template<typename X,typename Y>
	PlainPair& operator=(const PlainPair<X,Y>& other)
	{
	    first=other.first;
	    second=other.second;
	}
	// convert to std::pair
	operator std::pair<T,U>() const
	{
	    return std::pair<T,U>(first,second);
	}
    };

    // An OutputPair holds a pair of references to iterators, and writes to them on assignment
    template<typename Iter1,typename Iter2>
    struct OutputPair
    {
    private:
	Iter1& firstIter;
	Iter2& secondIter;

	OutputPair(const OutputPair&); // can't be copied
    public:
	OutputPair(Iter1& firstIter_,Iter2& secondIter_):
	    firstIter(firstIter_),secondIter(secondIter_)
	{}

	template<typename SomePair>
	OutputPair& operator=(const SomePair& other)
	{
	    *firstIter=other.first;
	    *secondIter=other.second;
	    return *this;
	}
    };

    // An OwningRefPair holds a pair of references, 
    // which may point to data within the pair, or external to it
    template<typename T,typename U>
    struct OwningRefPair
    {
    public:
	T& first;
	U& second;

	typedef T first_type;
	typedef U second_type;

    private:
	struct OwnedPair
	{
	    T first;
	    U second;
	    
	    template<typename Val1,typename Val2>
	    OwnedPair(Val1& v1,Val2& v2):
		first(v1),second(v2)
	    {}
	};
	
	utils::RawMem<OwnedPair> pairBuf;
	const bool ownsFlag;

	OwnedPair* getPairPtr()
	{
	    return reinterpret_cast<OwnedPair*>(pairBuf.data);
	}
	template<typename Val1,typename Val2>
	void createCopy(Val1& v1,Val2& v2)
	{
	    new(getPairPtr()) OwnedPair(v1,v2);
	}

    public:
	// copy from other types of pairs too
	template<typename SomePair>
	OwningRefPair(const SomePair& other):
	    first(getPairPtr()->first),
	    second(getPairPtr()->second),
	    ownsFlag(true)
	{
	    createCopy(other.first,other.second);
	}
	// copying copies values by default
	OwningRefPair(const OwningRefPair& other):
	    first(getPairPtr()->first),
	    second(getPairPtr()->second),
	    ownsFlag(true)
	{
	    createCopy(other.first,other.second);
	}

	// allow user to specify
	// whether to copy or preserve references
	template<typename SomePair>
	OwningRefPair(const SomePair& other,bool copy):
	    first(copy?getPairPtr()->first:other.first),
	    second(copy?getPairPtr()->second:other.second),
	    ownsFlag(copy)
	{
	    if(ownsFlag)
	    {
		createCopy(other.first,other.second);
	    }
	}

	// initialize with values
	template<typename Val1,typename Val2>
	OwningRefPair(const Val1& first_,const Val2& second_):
	    first(getPairPtr()->first),
	    second(getPairPtr()->second),
	    ownsFlag(true)
	{
	    createCopy(first_,second_);
	}

	// initialize with references
	OwningRefPair(T& first_,U& second_,bool copy):
	    first(copy?getPairPtr()->first:first_),
	    second(copy?getPairPtr()->second:second_),
	    ownsFlag(copy)
	{
	    if(ownsFlag)
	    {
		createCopy(first_,second_);
	    }
	}

	// if we own the values, destroy them
	~OwningRefPair()
	{
	    if(ownsFlag)
	    {
		getPairPtr()->~OwnedPair();
	    }
	}

	// assignment assigns to referenced values
	template<typename SomePair>
	OwningRefPair& operator=(const SomePair& other)
	{
	    first=other.first;
	    second=other.second;
	    return *this;
	}
	OwningRefPair& operator=(const OwningRefPair& other)
	{
	    first=other.first;
	    second=other.second;
	    return *this;
	}

	// convert to std::pair
	operator std::pair<T,U>() const
	{
	    return std::pair<T,U>(first,second);
	}
    
    };

    // create an owning pair
    template<typename T,typename U>
    OwningRefPair<T,U> makeOwningRefPair(const T& first_,const U& second_)
    {
	return PlainPair<T,U>(first_,second_);
    }

    namespace helper
    {
	// Define, construct and destroy the appropriate value_type for
	// the given iterator category
#ifdef NO_PARTIAL_SPECIALIZATION
	template<typename Category>
	struct ValueForCategory
	{
	    template<typename Iter1,typename Iter2>
	    struct Def
	    {
		typedef typename std::iterator_traits<Iter1>::value_type Value1;
		typedef typename std::iterator_traits<Iter2>::value_type Value2;
		typedef OwningRefPair<Value1,Value2> value_type;
		typedef value_type Type;

		static void construct(Type* p,Iter1& it1,Iter2& it2)
		{
		    // don't copy values, keep as references
		    new (p) Type(*it1,*it2,false);
		}
		
		static void destruct(Type* p)
		{
		    p->~OwningRefPair<Value1,Value2>();
		}
		
	    };
	};
    
	template<>
	struct ValueForCategory<std::input_iterator_tag>
	{
	    template<typename Iter1,typename Iter2>
	    struct Def
	    {
		typedef typename std::iterator_traits<Iter1>::value_type Value1;
		typedef typename std::iterator_traits<Iter2>::value_type Value2;
		typedef PlainPair<Value1,Value2> value_type;
		typedef value_type Type;

		static void construct(Type* p,Iter1& it1,Iter2& it2)
		{
		    new (p) Type(*it1,*it2);
		}
		
		static void destruct(Type* p)
		{
		    p->~PlainPair<Value1,Value2>();
		}
	    };
	};

	template<>
	struct ValueForCategory<std::output_iterator_tag>
	{
	    template<typename Iter1,typename Iter2>
	    struct Def
	    {
		typedef void value_type;
		typedef OutputPair<Iter1,Iter2> Type;

		static void construct(Type* p,Iter1& it1,Iter2& it2)
		{
		    new (p) Type(it1,it2);
		}
		
		static void destruct(Type* p)
		{
		    p->~OutputPair<Iter1,Iter2>();
		}
	    };
	};

	template<typename Category,typename Iter1,typename Iter2>
	struct VFCSelector
	{
	    typedef ValueForCategory<Category>::Def<Iter1,Iter2> Type;
	};
	
#else
	template<typename Category,typename Iter1,typename Iter2>
	struct ValueForCategory
	{
	    typedef typename std::iterator_traits<Iter1>::value_type Value1;
	    typedef typename std::iterator_traits<Iter2>::value_type Value2;
	    typedef OwningRefPair<Value1,Value2> value_type;
	    typedef value_type Type;
	    
	    static void construct(Type* p,Iter1& it1,Iter2& it2)
	    {
		// don't copy values, keep as references
		new (p) Type(*it1,*it2,false);
	    }
	    
	    static void destruct(Type* p)
	    {
		p->~OwningRefPair<Value1,Value2>();
	    }
	    
	};

	template<typename Iter1,typename Iter2>
	struct ValueForCategory<std::input_iterator_tag,Iter1,Iter2>
	{
	    typedef typename std::iterator_traits<Iter1>::value_type Value1;
	    typedef typename std::iterator_traits<Iter2>::value_type Value2;
	    typedef PlainPair<Value1,Value2> value_type;
	    typedef value_type Type;
	    
	    static void construct(Type* p,Iter1& it1,Iter2& it2)
	    {
		new (p) Type(*it1,*it2);
	    }
	    
	    static void destruct(Type* p)
	    {
		p->~PlainPair<Value1,Value2>();
	    }
	};

	template<typename Iter1,typename Iter2>
	struct ValueForCategory<std::output_iterator_tag,Iter1,Iter2>
	{
	    typedef void value_type;
	    typedef OutputPair<Iter1,Iter2> Type;
	    
	    static void construct(Type* p,Iter1& it1,Iter2& it2)
	    {
		new (p) Type(it1,it2);
	    }
	    
	    static void destruct(Type* p)
	    {
		p->~OutputPair<Iter1,Iter2>();
	    }
	};

	template<typename Category,typename Iter1,typename Iter2>
	struct VFCSelector
	{
	    typedef ValueForCategory<Category,Iter1,Iter2> Type;
	};
#endif

	

	// Select the iterator_category and value_type for our PairIt
	template<typename Iter1,typename Iter2>
	struct PairItHelper
	{
	    typedef typename std::iterator_traits<Iter1>::iterator_category Cat1;
	    typedef typename std::iterator_traits<Iter2>::iterator_category Cat2;

	    typedef typename CommonCategory<Cat1,Cat2>::Type iterator_category;
	    typedef typename VFCSelector<iterator_category,Iter1,Iter2>::Type ValueTypeDef;
	    typedef typename ValueTypeDef::value_type value_type;
	    typedef typename ValueTypeDef::Type DeRefType;

	    typedef DeRefType& reference;
	    typedef DeRefType* pointer;

	    typedef std::ptrdiff_t difference_type;

	    typedef std::iterator<iterator_category,value_type,difference_type,pointer,reference> IteratorType;

	    static void construct(DeRefType* p,Iter1& it1,Iter2& it2)
	    {
		ValueTypeDef::construct(p,it1,it2);
	    }
	    
	    static void destruct(DeRefType* p)
	    {
		ValueTypeDef::destruct(p);
	    }
	};
    }

    // the actual Pair Iterator itself
    template<typename Iter1,typename Iter2>
    struct PairIt:
	public helper::PairItHelper<Iter1,Iter2>::IteratorType
    {
    private:
	typedef helper::PairItHelper<Iter1,Iter2> PairDefs;
    public:
	typedef typename PairDefs::iterator_category iterator_category;
	typedef typename PairDefs::value_type value_type;
	typedef typename PairDefs::difference_type difference_type;
	typedef typename PairDefs::reference reference;
	typedef typename PairDefs::pointer pointer;
    private:
	pointer getValuePtr() const
	{
	    return reinterpret_cast<pointer>(dataCache.data);
	}
    
	void emptyCache() const
	{
	    if(cacheInitialized)
	    {
		PairDefs::destruct(getValuePtr());
		cacheInitialized=false;
	    }
	}

	void initCache() const
	{
	    emptyCache();
	    PairDefs::construct(getValuePtr(),it1,it2);
	    cacheInitialized=true;
	}
    
    
    public:
    
	PairIt(Iter1 it1_,Iter2 it2_):
	    it1(it1_),it2(it2_),cacheInitialized(false)
	{}
	template<typename I1,typename I2>
	PairIt(const PairIt<I1,I2>& other):
	    it1(other.it1),it2(other.it2),cacheInitialized(false)
	{}
	PairIt(const PairIt& other):
	    it1(other.it1),it2(other.it2),cacheInitialized(false)
	{}
	PairIt():
	    it1(),it2(),cacheInitialized(false)
	{}
    

	~PairIt()
	{
	    emptyCache();
	}

	void swap(PairIt& other)
	{
	    using std::swap;
	
	    swap(it1,other.it1);
	    swap(it2,other.it2);
	}

	template<typename I1,typename I2>
	void iter_swap(const PairIt<I1,I2>& other) const
	{
	    using std::iter_swap;
	
	    iter_swap(it1,other.it1);
	    iter_swap(it2,other.it2);
	}
    

	// Input Iterator requirements
	reference operator*() const
	{
	    initCache();
	    return *getValuePtr();
	}

	pointer operator->() const
	{
	    initCache();
	    return getValuePtr();
	}

	friend bool operator==(const PairIt& lhs,const PairIt& rhs)
	{
	    return lhs.it1==rhs.it1 && lhs.it2==rhs.it2;
	}

	friend bool operator!=(const PairIt& lhs,const PairIt& rhs)
	{
	    return lhs.it1!=rhs.it1 || lhs.it2!=rhs.it2;
	}

	// Forward Iterator requirements
	PairIt& operator++()
	{
	    ++it1;
	    ++it2;
	    return *this;
	}
    
	PairIt operator++(int)
	{
	    PairIt temp(*this);
	    ++*this;
	    return temp;
	}

	// Bidirectional Iterator requirements
	PairIt& operator--()
	{
	    --it1;
	    --it2;
	    return *this;
	}
    
	PairIt operator--(int)
	{
	    PairIt temp(*this);
	    --*this;
	    return temp;
	}
    
	// Random-Access Iterator requirements
	PairIt& operator+=(difference_type n)
	{
	    it1+=n;
	    it2+=n;
	    return *this;
	}

	PairIt& operator-=(difference_type n)
	{
	    it1-=n;
	    it2-=n;
	    return *this;
	}

	friend difference_type operator-(const PairIt& a,const PairIt& b)
	{
	    difference_type res1=a.it1-b.it1;
	    difference_type res2=a.it2-b.it2;
	
	    if(res1==res2)
	    {
		return res1;
	    }
	
	    throw std::logic_error("The iterators in the pairs are mismatched");
	}

	PairIt& operator[](difference_type n) const
	{
	    return *(*this+n);
	}

	// allow the user to extract the underlying iterators if required.
	Iter1 getFirstIterator() const
	{
	    return it1;
	}
	
	Iter2 getSecondIterator() const
	{
	    return it2;
	}

    private:
	// everything is mutable so we can modify it without affecting const correctness
	// of client code
	mutable Iter1 it1;
	mutable Iter2 it2;    
	mutable utils::RawMem<typename PairDefs::DeRefType> dataCache;
	mutable bool cacheInitialized;
    };

    // more random-access iterator requirements
    template<typename I1,typename I2>
    PairIt<I1,I2> operator+(std::ptrdiff_t n,PairIt<I1,I2> temp)
    {
	temp+=n;
	return temp;
    }

    template<typename I1,typename I2>
    PairIt<I1,I2> operator+(PairIt<I1,I2> temp,std::ptrdiff_t n)
    {
	temp+=n;
	return temp;
    }

    template<typename I1,typename I2>
    PairIt<I1,I2> operator-(PairIt<I1,I2> temp,std::ptrdiff_t n)
    {
	temp-=n;
	return temp;
    }

    template<typename I1,typename I2,typename I3,typename I4>
    bool operator<(const PairIt<I1,I2>& a,const PairIt<I3,I4>& b)
    {
	return (b-a)>0;
    }

    template<typename I1,typename I2,typename I3,typename I4>
    bool operator>(const PairIt<I1,I2>& a,const PairIt<I3,I4>& b)
    {
	return b<a;
    }

    template<typename I1,typename I2,typename I3,typename I4>
    bool operator>=(const PairIt<I1,I2>& a,const PairIt<I3,I4>& b)
    {
	return !(b<a);
    }

    template<typename I1,typename I2,typename I3,typename I4>
    bool operator<=(const PairIt<I1,I2>& a,const PairIt<I3,I4>& b)
    {
	return !(b>a);
    }

    // implementation of swap and iter_swap
    template<typename I1,typename I2>
    void swap(PairIt<I1,I2>& lhs,PairIt<I1,I2>& rhs)
    {
	lhs.swap(rhs);
    }
    
    template<typename I1,typename I2,typename I3,typename I4>
    void iter_swap(const PairIt<I1,I2>& lhs,const PairIt<I3,I4>& rhs)
    {
	lhs.iter_swap(rhs);
    }

    template<typename Iter1,typename Iter2>
    PairIt<Iter1,Iter2> makePairIterator(Iter1 i1,Iter2 i2)
    {
	return PairIt<Iter1,Iter2>(i1,i2);
    }
    
}

#endif
